from __future__ import annotations

from dataclasses import dataclass
from dataclasses import field
from datetime import datetime
from enum import Enum
from typing import Dict
from typing import List
from uuid import UUID


class MetadataValueType(str, Enum):
    LOCATION = "LOCATION"  # string
    STRING = "STRING"  # string
    LINK = "LINK"  # string
    BOOLEAN = "BOOLEAN"  # bool
    NUMBER = "NUMBER"  # float
    DATE = "DATE"  # datetime


@dataclass(frozen=True)
class MetadataValue:
    value: str
    type_: MetadataValueType


class FileState(str, Enum):
    OK = "OK"
    CORRUPTED = "CORRUPTED"
    UPLOADING = "UPLOADING"
    ERROR = "ERROR"
    CONVERSION_ERROR = "CONVERSION_ERROR"
    LOST = "LOST"
    FOUND = "FOUND"


@dataclass(frozen=True)
class Project:
    id: UUID
    name: str
    description: str
    created_at: datetime
    updated_at: datetime


@dataclass(frozen=True)
class Mission:
    id: UUID
    name: str
    created_at: datetime
    updated_at: datetime
    project_id: UUID
    project_name: str
    metadata: Dict[str, MetadataValue] = field(default_factory=dict)
    number_of_files: int = 0
    size: int = 0


@dataclass(frozen=True)
class File:
    id: UUID
    name: str
    hash: str
    size: int
    type_: str
    date: datetime
    created_at: datetime
    updated_at: datetime
    mission_id: UUID
    mission_name: str
    project_id: UUID
    project_name: str
    categories: List[str] = field(default_factory=list)
    topics: List[str] = field(default_factory=list)
    state: FileState = FileState.OK


# this is the file state for the verify command
class FileVerificationStatus(str, Enum):
    UPLAODED = "uploaded"
    UPLOADING = "uploading"
    COMPUTING_HASH = "computing hash"
    MISSING = "missing"
    MISMATCHED_HASH = "hash mismatch"
    UNKNOWN = "unknown"



from __future__ import annotations

from pathlib import Path
from typing import List
from typing import Optional

import typer

import kleinkram.core
import kleinkram.utils
from kleinkram.api.client import AuthenticatedClient
from kleinkram.api.query import MissionQuery
from kleinkram.api.query import ProjectQuery
from kleinkram.config import get_shared_state
from kleinkram.errors import FileNameNotSupported
from kleinkram.errors import MissionNotFound
from kleinkram.utils import load_metadata
from kleinkram.utils import split_args

HELP = """\
Upload files to kleinkram.
"""

upload_typer = typer.Typer(
    name="upload",
    no_args_is_help=True,
    invoke_without_command=True,
    help=HELP,
)


@upload_typer.callback()
def upload(
    files: List[str] = typer.Argument(help="files to upload"),
    project: Optional[str] = typer.Option(
        None, "--project", "-p", help="project id or name"
    ),
    mission: str = typer.Option(..., "--mission", "-m", help="mission id or name"),
    create: bool = typer.Option(False, help="create mission if it does not exist"),
    metadata: Optional[str] = typer.Option(
        None, help="path to metadata file (json or yaml)"
    ),
    fix_filenames: bool = typer.Option(
        False,
        help="fix filenames before upload, this does not change the filenames locally",
    ),
    ignore_missing_tags: bool = typer.Option(False, help="ignore mission tags"),
) -> None:
    # get filepaths
    file_paths = [Path(file) for file in files]

    mission_ids, mission_patterns = split_args([mission])
    project_ids, project_patterns = split_args([project] if project else [])

    project_query = ProjectQuery(ids=project_ids, patterns=project_patterns)
    mission_query = MissionQuery(
        ids=mission_ids,
        patterns=mission_patterns,
        project_query=project_query,
    )

    if not fix_filenames:
        for file in file_paths:
            if not kleinkram.utils.check_filename_is_sanatized(file.stem):
                raise FileNameNotSupported(
                    f"Only `{''.join(kleinkram.utils.INTERNAL_ALLOWED_CHARS)}` are "
                    f"allowed in filenames and at most 50 chars: {file}. "
                    f"Consider using `--fix-filenames`"
                )

    try:
        kleinkram.core.upload(
            client=AuthenticatedClient(),
            query=mission_query,
            file_paths=file_paths,
            create=create,
            metadata=load_metadata(Path(metadata)) if metadata else None,
            ignore_missing_metadata=ignore_missing_tags,
            verbose=get_shared_state().verbose,
        )
    except MissionNotFound:
        if create:
            raise  # dont change the error message
        raise MissionNotFound("Mission not found. Use `--create` to create it.")



from __future__ import annotations

import datetime

import pytest
from rich.table import Table

from kleinkram.models import MetadataValue
from kleinkram.models import MetadataValueType
from kleinkram.printing import _add_placeholder_row
from kleinkram.printing import format_bytes
from kleinkram.printing import parse_metadata_value


def test_format_bytes():
    assert format_bytes(0) == "0 B"
    assert format_bytes(1) == "1 B"
    assert format_bytes(1000) == "1000 B"
    assert format_bytes(1024) == "1.00 KB"
    assert format_bytes(1025) == "1.00 KB"
    assert format_bytes(2048) == "2.00 KB"
    assert format_bytes(2**20) == "1.00 MB"
    assert format_bytes(2**30) == "1.00 GB"
    assert format_bytes(2**40) == "1.00 TB"
    assert format_bytes(2**50) == "1.00 PB"


def test_add_placeholder_row():
    table = Table("foo", "bar")
    _add_placeholder_row(table, skipped=1)

    assert table.row_count == 1
    assert table.columns[0]._cells[-1] == "... (1 more)"
    assert table.columns[1]._cells[-1] == "..."


def test_parse_metadata_value():
    mv = MetadataValue(type_=MetadataValueType.STRING, value="foo")
    assert parse_metadata_value(mv) == "foo"
    mv = MetadataValue(type_=MetadataValueType.LINK, value="foo")
    assert parse_metadata_value(mv) == "foo"
    mv = MetadataValue(type_=MetadataValueType.LOCATION, value="foo")
    assert parse_metadata_value(mv) == "foo"
    mv = MetadataValue(type_=MetadataValueType.NUMBER, value="1")
    assert parse_metadata_value(mv) == 1.0
    mv = MetadataValue(type_=MetadataValueType.BOOLEAN, value="true")
    assert parse_metadata_value(mv) is True  # noqa
    mv = MetadataValue(type_=MetadataValueType.BOOLEAN, value="false")
    assert parse_metadata_value(mv) is False  # noqa
    mv = MetadataValue(type_=MetadataValueType.DATE, value="2021-01-01T00:00:00Z")
    assert parse_metadata_value(mv) == datetime.datetime(
        2021, 1, 1, 0, 0, 0, tzinfo=datetime.timezone.utc
    )


@pytest.mark.skip
def test_projects_to_table(): ...


@pytest.mark.skip
def test_missions_to_table(): ...


@pytest.mark.skip
def test_files_to_table(): ...


@pytest.mark.skip
def test_mission_info_table(): ...


@pytest.mark.skip
def test_project_info_table(): ...



from __future__ import annotations

import os
from pathlib import Path
from tempfile import TemporaryDirectory
from unittest import mock

import pytest

import kleinkram.config
from kleinkram.config import ACTION_API
from kleinkram.config import ACTION_API_KEY
from kleinkram.config import ACTION_S3
from kleinkram.config import Config
from kleinkram.config import Endpoint
from kleinkram.config import _load_config
from kleinkram.config import _load_config_if_compatible
from kleinkram.config import add_endpoint
from kleinkram.config import check_config_compatibility
from kleinkram.config import endpoint_table
from kleinkram.config import get_config
from kleinkram.config import get_env
from kleinkram.config import get_shared_state
from kleinkram.config import save_config
from kleinkram.config import select_endpoint

CONFIG_FILENAME = ".kleinkram.json"


TEST_API_KEY = "test_key"
TEST_API = "test_api"
TEST_S3 = "test_s3"


@pytest.fixture()
def set_api_key_env(monkeypatch):
    with mock.patch.dict(os.environ, clear=True):
        envvars = {ACTION_API_KEY: TEST_API_KEY}
        for k, v in envvars.items():
            monkeypatch.setenv(k, v)
        yield  # This is the magical bit which restore the environment after


@pytest.fixture()
def set_endpoint_env(monkeypatch):
    with mock.patch.dict(os.environ, clear=True):
        envvars = {ACTION_API: TEST_API, ACTION_S3: TEST_S3}
        for k, v in envvars.items():
            monkeypatch.setenv(k, v)
        yield  # This is the magical bit which restore the environment after


@pytest.fixture
def config_path():
    with TemporaryDirectory() as tmpdir:
        yield Path(tmpdir) / CONFIG_FILENAME


def test_load_config_if_compatible_with_invalid_config(config_path):
    with open(config_path, "w") as f:
        f.write("this is not a valid config")
    assert _load_config_if_compatible(config_path) is None


def test_load_config_default(config_path):
    config = _load_config(path=config_path)

    assert not config_path.exists()
    assert Config() == config

    assert config.endpoint_credentials == {}
    assert config.selected_endpoint == get_env().value


def test_load_default_config_with_env_var_api_key_specified(
    config_path, set_api_key_env
):
    assert set_api_key_env is None

    config = _load_config(path=config_path)

    creds = config.endpoint_credentials[config.selected_endpoint]
    assert creds.auth_token is None
    assert creds.refresh_token is None
    assert creds.api_key == TEST_API_KEY

    assert not config_path.exists()


def test_load_default_config_with_env_var_endpoints_specified(
    config_path, set_endpoint_env
):
    assert set_endpoint_env is None
    config = _load_config(path=config_path)

    assert config.selected_endpoint == "action"
    assert config.endpoint == Endpoint("action", TEST_API, TEST_S3)

    assert not config_path.exists()


def test_save_and_load_config(config_path):
    config = Config(version="foo")

    assert not config_path.exists()
    with mock.patch.object(kleinkram.config.logger, "warning") as mock_warning:
        save_config(config, path=config_path)
        mock_warning.assert_not_called()

    assert config_path.exists()
    loaded_config = _load_config(path=config_path)
    assert loaded_config == config


def test_save_and_load_config_when_tmpfile_fails(config_path):
    config = Config(version="foo")

    assert not config_path.exists()
    with mock.patch("tempfile.mkstemp", side_effect=Exception), mock.patch.object(
        kleinkram.config.logger, "warning"
    ) as mock_warning:
        save_config(config, path=config_path)
        mock_warning.assert_called_once()

    assert config_path.exists()
    loaded_config = _load_config(path=config_path)
    assert loaded_config == config


def test_get_config_default(config_path):
    config = get_config(path=config_path)

    assert not config_path.exists()
    assert Config() == config
    assert config is get_config(path=config_path)


def test_get_config_after_save(config_path):
    config = get_config(path=config_path)
    config.version = "foo"
    save_config(config, path=config_path)

    assert config is get_config(path=config_path)


def test_get_shared_state():
    state = get_shared_state()
    assert state is get_shared_state()


def test_select_endpoint(config_path):
    config = get_config(path=config_path)
    save_config(config, path=config_path)
    assert config.selected_endpoint == get_env().value

    # select existing endpoint
    select_endpoint(config, "prod", path=config_path)
    assert config.selected_endpoint == "prod"
    assert config == _load_config(path=config_path)

    with pytest.raises(ValueError):
        select_endpoint(config, "foo", path=config_path)


def test_add_endpoint(config_path):
    config = get_config(path=config_path)
    save_config(config, path=config_path)
    assert config.selected_endpoint == get_env().value

    with pytest.raises(ValueError):
        select_endpoint(config, "foo", path=config_path)

    ep = Endpoint("foo", "api", "s3")
    add_endpoint(config, ep, path=config_path)
    assert config.selected_endpoint == "foo"
    assert config.endpoint == ep
    assert config == _load_config(path=config_path)

    select_endpoint(config, "dev", path=config_path)
    assert config.selected_endpoint == "dev"
    select_endpoint(config, "foo", path=config_path)
    assert config.selected_endpoint == "foo"


def test_endpoint_table():
    config = Config()
    table = endpoint_table(config)

    assert [c.header for c in table.columns] == ["Name", "API", "S3"]
    assert len(table.rows) == 3


def test_check_config_compatiblity(config_path):
    assert check_config_compatibility(path=config_path)
    with open(config_path, "w") as f:
        f.write("foo")  # invalid config
    assert not check_config_compatibility(path=config_path)



from __future__ import annotations

import logging
import sys
from concurrent.futures import Future
from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import as_completed
from enum import Enum
from pathlib import Path
from time import monotonic
from typing import Dict
from typing import NamedTuple
from typing import Optional
from typing import Tuple
from uuid import UUID

import boto3.s3.transfer
import botocore.config
import httpx
from rich.console import Console
from tqdm import tqdm

from kleinkram.api.client import AuthenticatedClient
from kleinkram.config import get_config
from kleinkram.errors import AccessDenied
from kleinkram.models import File
from kleinkram.models import FileState
from kleinkram.utils import b64_md5
from kleinkram.utils import format_error
from kleinkram.utils import format_traceback
from kleinkram.utils import styled_string

logger = logging.getLogger(__name__)

UPLOAD_CREDS = "/files/temporaryAccess"
UPLOAD_CONFIRM = "/queue/confirmUpload"
UPLOAD_CANCEL = "/files/cancelUpload"

DOWNLOAD_CHUNK_SIZE = 1024 * 1024 * 16
DOWNLOAD_URL = "/files/download"

S3_MAX_RETRIES = 60  # same as frontend
S3_READ_TIMEOUT = 60 * 5  # 5 minutes


class UploadCredentials(NamedTuple):
    access_key: str
    secret_key: str
    session_token: str
    file_id: UUID
    bucket: str


def _confirm_file_upload(
    client: AuthenticatedClient, file_id: UUID, file_hash: str
) -> None:
    data = {
        "uuid": str(file_id),
        "md5": file_hash,
    }
    resp = client.post(UPLOAD_CONFIRM, json=data)
    resp.raise_for_status()


def _cancel_file_upload(
    client: AuthenticatedClient, file_id: UUID, mission_id: UUID
) -> None:
    data = {
        "uuid": [str(file_id)],
        "missionUUID": str(mission_id),
    }
    resp = client.post(UPLOAD_CANCEL, json=data)
    resp.raise_for_status()
    return


FILE_EXISTS_ERROR = "File already exists"

# fields for upload credentials
ACCESS_KEY_FIELD = "accessKey"
SECRET_KEY_FIELD = "secretKey"
SESSION_TOKEN_FIELD = "sessionToken"
CREDENTIALS_FIELD = "accessCredentials"
FILE_ID_FIELD = "fileUUID"
BUCKET_FIELD = "bucket"


def _get_upload_creditials(
    client: AuthenticatedClient, internal_filename: str, mission_id: UUID
) -> Optional[UploadCredentials]:
    dct = {
        "filenames": [internal_filename],
        "missionUUID": str(mission_id),
    }
    resp = client.post(UPLOAD_CREDS, json=dct)
    resp.raise_for_status()

    data = resp.json()["data"][0]

    if data.get("error") == FILE_EXISTS_ERROR:
        return None

    bucket = data[BUCKET_FIELD]
    file_id = UUID(data[FILE_ID_FIELD], version=4)

    creds = data[CREDENTIALS_FIELD]
    access_key = creds[ACCESS_KEY_FIELD]
    secret_key = creds[SECRET_KEY_FIELD]
    session_token = creds[SESSION_TOKEN_FIELD]

    return UploadCredentials(
        access_key=access_key,
        secret_key=secret_key,
        session_token=session_token,
        file_id=file_id,
        bucket=bucket,
    )


def _s3_upload(
    local_path: Path,
    *,
    endpoint: str,
    credentials: UploadCredentials,
    pbar: tqdm,
) -> None:
    # configure boto3
    config = botocore.config.Config(
        retries={"max_attempts": S3_MAX_RETRIES},
        read_timeout=S3_READ_TIMEOUT,
    )
    client = boto3.client(
        "s3",
        endpoint_url=endpoint,
        aws_access_key_id=credentials.access_key,
        aws_secret_access_key=credentials.secret_key,
        aws_session_token=credentials.session_token,
        config=config,
    )
    client.upload_file(
        str(local_path),
        credentials.bucket,
        str(credentials.file_id),
        Callback=pbar.update,
    )


class UploadState(Enum):
    UPLOADED = 1
    EXISTS = 2
    CANCELED = 3


# TODO: i dont want to handle errors at this level
def upload_file(
    client: AuthenticatedClient,
    *,
    mission_id: UUID,
    filename: str,
    path: Path,
    verbose: bool = False,
    s3_endpoint: Optional[str] = None,
) -> UploadState:
    """\
    returns bytes uploaded
    """
    if s3_endpoint is None:
        s3_endpoint = get_config().endpoint.s3

    total_size = path.stat().st_size
    with tqdm(
        total=total_size,
        unit="B",
        unit_scale=True,
        desc=f"uploading {path}...",
        leave=False,
        disable=not verbose,
    ) as pbar:
        # get per file upload credentials
        creds = _get_upload_creditials(
            client, internal_filename=filename, mission_id=mission_id
        )
        if creds is None:
            return UploadState.EXISTS

        try:
            _s3_upload(path, endpoint=s3_endpoint, credentials=creds, pbar=pbar)
        except Exception as e:
            logger.error(format_traceback(e))
            _cancel_file_upload(client, creds.file_id, mission_id)
            return UploadState.CANCELED
        else:
            _confirm_file_upload(client, creds.file_id, b64_md5(path))
            return UploadState.UPLOADED


def _get_file_download(client: AuthenticatedClient, id: UUID) -> str:
    """\
    get the download url for a file by file id
    """
    resp = client.get(DOWNLOAD_URL, params={"uuid": str(id), "expires": True})

    if 400 <= resp.status_code < 500:
        raise AccessDenied(
            f"Failed to download file: {resp.json()['message']}"
            f"Status Code: {resp.status_code}",
        )

    resp.raise_for_status()

    return resp.text


def _url_download(
    url: str, *, path: Path, size: int, overwrite: bool = False, verbose: bool = False
) -> None:
    if path.exists() and not overwrite:
        raise FileExistsError(f"file already exists: {path}")

    with httpx.stream("GET", url) as response:
        with open(path, "wb") as f:
            with tqdm(
                total=size,
                desc=f"downloading {path.name}",
                unit="B",
                unit_scale=True,
                leave=False,
                disable=not verbose,
            ) as pbar:
                for chunk in response.iter_bytes(chunk_size=DOWNLOAD_CHUNK_SIZE):
                    f.write(chunk)
                    pbar.update(len(chunk))


class DownloadState(Enum):
    DOWNLOADED_OK = 1
    SKIPPED_OK = 2
    DOWNLOADED_INVALID_HASH = 3
    SKIPPED_INVALID_HASH = 4
    SKIPPED_INVALID_REMOTE_STATE = 5


def download_file(
    client: AuthenticatedClient,
    *,
    file: File,
    path: Path,
    overwrite: bool = False,
    verbose: bool = False,
) -> DownloadState:
    # skip files that are not ok on remote
    if file.state != FileState.OK:
        return DownloadState.SKIPPED_INVALID_REMOTE_STATE

    # skip existing files depending on flags set
    if path.exists():
        local_hash = b64_md5(path)
        if local_hash != file.hash and not overwrite and file.hash is not None:
            return DownloadState.SKIPPED_INVALID_HASH

        elif local_hash == file.hash:
            return DownloadState.SKIPPED_OK

        # this has to be here
        if verbose:
            tqdm.write(
                styled_string(f"overwriting {path}, hash missmatch", style="yellow")
            )

    # request a download url
    download_url = _get_file_download(client, file.id)

    # create parent directories
    path.parent.mkdir(parents=True, exist_ok=True)

    # download the file and check the hash
    _url_download(
        download_url, path=path, size=file.size, overwrite=overwrite, verbose=verbose
    )
    observed_hash = b64_md5(path)
    if file.hash is not None and observed_hash != file.hash:
        return DownloadState.DOWNLOADED_INVALID_HASH
    return DownloadState.DOWNLOADED_OK


UPLOAD_STATE_COLOR = {
    UploadState.UPLOADED: "green",
    UploadState.EXISTS: "yellow",
    UploadState.CANCELED: "red",
}


def _upload_handler(
    future: Future[UploadState], path: Path, *, verbose: bool = False
) -> int:
    try:
        state = future.result()
    except Exception as e:
        logger.error(format_traceback(e))
        if verbose:
            tqdm.write(format_error(f"error uploading {path}", e))
        else:
            print(path.absolute(), file=sys.stderr)
        return 0

    if state == UploadState.UPLOADED:
        msg = f"uploaded {path}"
    elif state == UploadState.EXISTS:
        msg = f"skipped {path} already uploaded"
    else:
        msg = f"canceled {path} upload"

    if verbose:
        tqdm.write(styled_string(msg, style=UPLOAD_STATE_COLOR[state]))
    else:
        stream = sys.stdout if state == UploadState.UPLOADED else sys.stderr
        print(path.absolute(), file=stream)

    return path.stat().st_size if state == UploadState.UPLOADED else 0


DOWNLOAD_STATE_COLOR = {
    DownloadState.DOWNLOADED_OK: "green",
    DownloadState.SKIPPED_OK: "green",
    DownloadState.DOWNLOADED_INVALID_HASH: "red",
    DownloadState.SKIPPED_INVALID_HASH: "yellow",
    DownloadState.SKIPPED_INVALID_REMOTE_STATE: "purple",
}


def _download_handler(
    future: Future[DownloadState], file: File, path: Path, *, verbose: bool = False
) -> int:
    try:
        state = future.result()
    except Exception as e:
        logger.error(format_traceback(e))
        if verbose:
            tqdm.write(format_error(f"error uploading {path}", e))
        else:
            print(path.absolute(), file=sys.stderr)
        return 0

    if state == DownloadState.DOWNLOADED_OK:
        msg = f"downloaded {path}"
    elif state == DownloadState.DOWNLOADED_INVALID_HASH:
        msg = f"downloaded {path} failed hash check"
    elif state == DownloadState.SKIPPED_OK:
        msg = f"skipped {path} already downloaded"
    elif state == DownloadState.SKIPPED_INVALID_HASH:
        msg = f"skipped {path} already downloaded, hash missmatch, cosider using `--overwrite`"
    else:
        msg = f"skipped {path} remote file has invalid state"

    if verbose:
        tqdm.write(styled_string(msg, style=DOWNLOAD_STATE_COLOR[state]))
    else:
        stream = (
            sys.stdout
            if state in (DownloadState.DOWNLOADED_OK, DownloadState.SKIPPED_OK)
            else sys.stderr
        )
        print(path.absolute(), file=stream)

    # number of bytes downloaded
    return file.size if state == DownloadState.DOWNLOADED_OK else 0


def upload_files(
    client: AuthenticatedClient,
    files: Dict[str, Path],
    mission_id: UUID,
    *,
    verbose: bool = False,
    n_workers: int = 2,
) -> None:
    with tqdm(
        total=len(files),
        unit="files",
        desc="uploading files",
        disable=not verbose,
        leave=False,
    ) as pbar:
        start = monotonic()
        futures: Dict[Future[UploadState], Path] = {}
        with ThreadPoolExecutor(max_workers=n_workers) as executor:
            for name, path in files.items():
                future = executor.submit(
                    upload_file,
                    client=client,
                    mission_id=mission_id,
                    filename=name,
                    path=path,
                    verbose=verbose,
                )
                futures[future] = path

            total_size = 0
            for future in as_completed(futures):
                size = _upload_handler(future, futures[future], verbose=verbose)
                total_size += size / 1024 / 1024

                pbar.update()
            pbar.refresh()

        t = monotonic() - start
        c = Console(file=sys.stderr)
        c.print(f"upload took {t:.2f} seconds")
        c.print(f"total size: {int(total_size)} MB")
        c.print(f"average speed: {total_size / t:.2f} MB/s")


def download_files(
    client: AuthenticatedClient,
    files: Dict[Path, File],
    *,
    verbose: bool = False,
    overwrite: bool = False,
    n_workers: int = 2,
) -> None:
    with tqdm(
        total=len(files),
        unit="files",
        desc="downloading files",
        disable=not verbose,
        leave=False,
    ) as pbar:

        start = monotonic()
        futures: Dict[Future[DownloadState], Tuple[File, Path]] = {}
        with ThreadPoolExecutor(max_workers=n_workers) as executor:
            for path, file in files.items():
                future = executor.submit(
                    download_file,
                    client=client,
                    file=file,
                    path=path,
                    overwrite=overwrite,
                    verbose=verbose,
                )
                futures[future] = (file, path)

            total_size = 0
            for future in as_completed(futures):
                file, path = futures[future]
                size = _download_handler(future, file, path, verbose=verbose)
                total_size += size / 1024 / 1024  # MB
                pbar.update()
            pbar.refresh()

        time = monotonic() - start
        c = Console(file=sys.stderr)
        c.print(f"download took {time:.2f} seconds")
        c.print(f"total size: {int(total_size)} MB")
        c.print(f"average speed: {total_size  / time:.2f} MB/s")



from __future__ import annotations

import logging
from collections import abc
from threading import Lock
from typing import Any
from typing import List
from typing import Mapping
from typing import Sequence
from typing import Tuple
from typing import Union

import httpx
from httpx._types import PrimitiveData

from kleinkram.config import Config
from kleinkram.config import Credentials
from kleinkram.config import get_config
from kleinkram.config import save_config
from kleinkram.errors import NotAuthenticated

logger = logging.getLogger(__name__)


COOKIE_AUTH_TOKEN = "authtoken"
COOKIE_REFRESH_TOKEN = "refreshtoken"
COOKIE_API_KEY = "clikey"


Data = Union[PrimitiveData, Any]
NestedData = Mapping[str, Data]
ListData = Sequence[Data]
QueryParams = Mapping[str, Union[Data, NestedData, ListData]]


def _convert_nested_data_query_params_values(
    key: str, values: NestedData
) -> List[Tuple[str, Data]]:
    return [(f"{key}[{k}]", v) for k, v in values.items()]


def _convert_list_data_query_params_values(
    key: str, values: ListData
) -> List[Tuple[str, Data]]:
    return [(key, value) for value in values]


def _convert_query_params_to_httpx_format(
    params: QueryParams,
) -> List[Tuple[str, Data]]:
    ret: List[Tuple[str, Data]] = []
    for key, value in params.items():
        if isinstance(value, (str, int, float, bool)) or value is None:
            ret.append((key, value))
        elif isinstance(value, abc.Mapping):
            ret.extend(_convert_nested_data_query_params_values(key, value))
        elif isinstance(value, abc.Sequence):
            ret.extend(_convert_list_data_query_params_values(key, value))
        else:  # TODO: handle this better
            ret.append((key, str(value)))
    return ret


class AuthenticatedClient(httpx.Client):
    _config: Config
    _config_lock: Lock

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

        self._config = get_config()
        self._config_lock = Lock()

        if self._config.credentials is None:
            logger.info("not authenticated...")
            raise NotAuthenticated
        elif (api_key := self._config.credentials.api_key) is not None:
            logger.info("using cli key...")
            self.cookies.set(COOKIE_API_KEY, api_key)
        else:
            logger.info("using refresh token...")
            assert self._config.credentials.auth_token is not None, "unreachable"
            self.cookies.set(COOKIE_AUTH_TOKEN, self._config.credentials.auth_token)

    def _refresh_token(self) -> None:
        if self._config.credentials is None:
            raise NotAuthenticated

        if self._config.credentials.api_key is not None:
            raise RuntimeError("cannot refresh token when using cli key auth")

        refresh_token = self._config.credentials.refresh_token
        if refresh_token is None:
            raise RuntimeError("no refresh token found")
        self.cookies.set(COOKIE_REFRESH_TOKEN, refresh_token)

        logger.info("refreshing token...")
        response = self.post(
            "/auth/refresh-token",
        )
        response.raise_for_status()
        new_access_token = response.cookies[COOKIE_AUTH_TOKEN]
        creds = Credentials(auth_token=new_access_token, refresh_token=refresh_token)

        logger.info("saving new tokens...")

        with self._config_lock:
            self._config.credentials = creds
            save_config(self._config)

        self.cookies.set(COOKIE_AUTH_TOKEN, new_access_token)

    def request(
        self,
        method: str,
        url: str | httpx.URL,
        params: QueryParams | None = None,
        *args: Any,
        **kwargs: Any,
    ) -> httpx.Response:
        if isinstance(url, httpx.URL):
            raise NotImplementedError(f"`httpx.URL` is not supported {url!r}")
        if not url.startswith("/"):
            url = f"/{url}"

        # try to do a request
        full_url = f"{self._config.endpoint.api}{url}"
        logger.info(f"requesting {method} {full_url}")

        httpx_params = _convert_query_params_to_httpx_format(params or {})
        response = super().request(
            method, full_url, params=httpx_params, *args, **kwargs
        )

        logger.info(f"got response {response}")

        # if the requesting a refresh token fails, we are not logged in
        if (url == "/auth/refresh-token") and response.status_code == 401:
            logger.info("got 401, not logged in...")
            raise NotAuthenticated

        # otherwise we try to refresh the token
        if response.status_code == 401:
            logger.info("got 401, trying to refresh token...")
            try:
                self._refresh_token()
            except Exception:
                raise NotAuthenticated

            logger.info(f"retrying request {method} {full_url}")
            resp = super().request(
                method, full_url, params=httpx_params, *args, **kwargs
            )
            logger.info(f"got response {resp}")
            return resp
        else:
            return response



from __future__ import annotations

import json
from enum import Enum
from typing import Any
from typing import Dict
from typing import Generator
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple
from uuid import UUID

import httpx

import kleinkram.errors
from kleinkram.api.client import AuthenticatedClient
from kleinkram.api.deser import FileObject
from kleinkram.api.deser import MissionObject
from kleinkram.api.deser import ProjectObject
from kleinkram.api.deser import _parse_file
from kleinkram.api.deser import _parse_mission
from kleinkram.api.deser import _parse_project
from kleinkram.api.pagination import paginated_request
from kleinkram.api.query import FileQuery
from kleinkram.api.query import MissionQuery
from kleinkram.api.query import ProjectQuery
from kleinkram.api.query import file_query_is_unique
from kleinkram.api.query import mission_query_is_unique
from kleinkram.api.query import project_query_is_unique
from kleinkram.config import get_config
from kleinkram.errors import AccessDenied
from kleinkram.errors import InvalidFileQuery
from kleinkram.errors import InvalidMissionMetadata
from kleinkram.errors import InvalidMissionQuery
from kleinkram.errors import InvalidProjectQuery
from kleinkram.errors import MissionExists
from kleinkram.errors import MissionNotFound
from kleinkram.errors import ProjectExists
from kleinkram.errors import ProjectNotFound
from kleinkram.models import File
from kleinkram.models import Mission
from kleinkram.models import Project
from kleinkram.utils import is_valid_uuid4

__all__ = [
    "_get_api_version",
    "_claim_admin",
    "_create_mission",
    "_create_project",
    "_update_mission",
    "_update_project",
    "_delete_files",
    "_delete_mission",
    "_delete_project",
    "get_files",
    "get_missions",
    "get_projects",
    "get_project",
    "get_mission",
    "get_file",
]


CLAIM_ADMIN = "/user/claimAdmin"
GET_STATUS = "/user/me"

UPDATE_PROJECT = "/projects"
UPDATE_MISSION = "/missions/tags"  # TODO: just metadata for now
CREATE_MISSION = "/missions/create"
CREATE_PROJECT = "/projects"


FILE_ENDPOINT = "/files"
MISSION_ENDPOINT = "/missions"
PROJECT_ENDPOINT = "/projects"

TAG_TYPE_BY_NAME = "/tag/filtered"


class Params(str, Enum):
    FILE_PATTERNS = "filePatterns"
    FILE_IDS = "fileUuids"
    MISSION_PATTERNS = "missionPatterns"
    MISSION_IDS = "missionUuids"
    PROJECT_PATTERNS = "projectPatterns"
    PROJECT_IDS = "projectUuids"


def _handle_list_params(params: Dict[str, Any]) -> Dict[str, Any]:
    """
    json dumps lists
    """
    new_params = {}
    for k, v in params.items():
        if not isinstance(v, list):
            new_params[k] = v
        else:
            new_params[k] = json.dumps(v)
    return new_params


def _project_query_to_params(
    project_query: ProjectQuery,
) -> Dict[str, List[str]]:
    params = {}
    if project_query.patterns:
        params[Params.PROJECT_PATTERNS.value] = project_query.patterns
    if project_query.ids:
        params[Params.PROJECT_IDS.value] = list(map(str, project_query.ids))
    return params


def _mission_query_to_params(mission_query: MissionQuery) -> Dict[str, List[str]]:
    params = _project_query_to_params(mission_query.project_query)
    if mission_query.patterns:
        params[Params.MISSION_PATTERNS.value] = mission_query.patterns
    if mission_query.ids:
        params[Params.MISSION_IDS.value] = list(map(str, mission_query.ids))
    return params


def _file_query_to_params(file_query: FileQuery) -> Dict[str, List[str]]:
    params = _mission_query_to_params(file_query.mission_query)
    if file_query.patterns:
        params[Params.FILE_PATTERNS.value] = list(file_query.patterns)
    if file_query.ids:
        params[Params.FILE_IDS.value] = list(map(str, file_query.ids))
    return params


def get_files(
    client: AuthenticatedClient,
    file_query: FileQuery,
    max_entries: Optional[int] = None,
) -> Generator[File, None, None]:
    params = _file_query_to_params(file_query)
    response_stream = paginated_request(
        client, FILE_ENDPOINT, params=params, max_entries=max_entries
    )
    yield from map(lambda f: _parse_file(FileObject(f)), response_stream)


def get_missions(
    client: AuthenticatedClient,
    mission_query: MissionQuery,
    max_entries: Optional[int] = None,
) -> Generator[Mission, None, None]:
    params = _mission_query_to_params(mission_query)
    response_stream = paginated_request(
        client, MISSION_ENDPOINT, params=params, max_entries=max_entries
    )
    yield from map(lambda m: _parse_mission(MissionObject(m)), response_stream)


def get_projects(
    client: AuthenticatedClient,
    project_query: ProjectQuery,
    max_entries: Optional[int] = None,
) -> Generator[Project, None, None]:
    params = _project_query_to_params(project_query)
    response_stream = paginated_request(
        client, PROJECT_ENDPOINT, params=params, max_entries=max_entries
    )
    yield from map(lambda p: _parse_project(ProjectObject(p)), response_stream)


def get_project(client: AuthenticatedClient, query: ProjectQuery) -> Project:
    """\
    get a unique project by specifying a project spec
    """
    if not project_query_is_unique(query):
        raise InvalidProjectQuery(
            f"Project query does not uniquely determine project: {query}"
        )
    try:
        return next(get_projects(client, query))
    except StopIteration:
        raise ProjectNotFound(f"Project not found: {query}")


def get_mission(client: AuthenticatedClient, query: MissionQuery) -> Mission:
    """\
    get a unique mission by specifying a mission query
    """
    if not mission_query_is_unique(query):
        raise InvalidMissionQuery(
            f"Mission query does not uniquely determine mission: {query}"
        )
    try:
        return next(get_missions(client, query))
    except StopIteration:
        raise MissionNotFound(f"Mission not found: {query}")


def get_file(client: AuthenticatedClient, query: FileQuery) -> File:
    """\
    get a unique file by specifying a file query
    """
    if not file_query_is_unique(query):
        raise InvalidFileQuery(f"File query does not uniquely determine file: {query}")
    try:
        return next(get_files(client, query))
    except StopIteration:
        raise kleinkram.errors.FileNotFound(f"File not found: {query}")


def _mission_name_is_available(
    client: AuthenticatedClient, mission_name: str, project_id: UUID
) -> bool:
    mission_query = MissionQuery(
        patterns=[mission_name], project_query=ProjectQuery(ids=[project_id])
    )
    try:
        _ = get_mission(client, mission_query)
    except MissionNotFound:
        return True
    return False


def _project_name_is_available(client: AuthenticatedClient, project_name: str) -> bool:
    project_query = ProjectQuery(patterns=[project_name])
    try:
        _ = get_project(client, project_query)
    except ProjectNotFound:
        return True
    return False


def _create_mission(
    client: AuthenticatedClient,
    project_id: UUID,
    mission_name: str,
    *,
    metadata: Optional[Dict[str, str]] = None,
    ignore_missing_tags: bool = False,
) -> UUID:
    """\
    creates a new mission with the given name and project_id

    if check_exists is True, the function will return the existing mission_id,
    otherwise if the mission already exists an error will be raised
    """
    if metadata is None:
        metadata = {}

    if not _mission_name_is_available(client, mission_name, project_id):
        raise MissionExists(
            f"Mission with name: `{mission_name}` already exists"
            f" in project: {project_id}"
        )

    if is_valid_uuid4(mission_name):
        raise ValueError(
            f"Mission name: `{mission_name}` is a valid UUIDv4, "
            "mission names must not be valid UUIDv4's"
        )

    # we need to translate tag keys to tag type ids
    tags = _get_tags_map(client, metadata)

    payload = {
        "name": mission_name,
        "projectUUID": str(project_id),
        "tags": {str(k): v for k, v in tags.items()},
        "ignoreTags": ignore_missing_tags,
    }

    resp = client.post(CREATE_MISSION, json=payload)
    resp.raise_for_status()

    return UUID(resp.json()["uuid"], version=4)


def _create_project(
    client: AuthenticatedClient, project_name: str, description: str
) -> UUID:
    if not _project_name_is_available(client, project_name):
        raise ProjectExists(f"Project with name: `{project_name}` already exists")

    # TODO: check name and description are valid
    payload = {"name": project_name, "description": description}
    resp = client.post(CREATE_PROJECT, json=payload)
    resp.raise_for_status()

    return UUID(resp.json()["uuid"], version=4)


def _get_metadata_type_id_by_name(
    client: AuthenticatedClient, tag_name: str
) -> Optional[UUID]:
    resp = client.get(TAG_TYPE_BY_NAME, params={"name": tag_name, "take": 1})

    if resp.status_code in (403, 404):
        return None

    resp.raise_for_status()

    data = resp.json()[0]
    return UUID(data["uuid"], version=4)


def _get_tags_map(
    client: AuthenticatedClient, metadata: Dict[str, str]
) -> Dict[UUID, str]:
    # TODO: this needs a better endpoint
    # why are we using metadata type ids as keys???
    ret = {}
    for key, val in metadata.items():
        metadata_type_id = _get_metadata_type_id_by_name(client, key)
        if metadata_type_id is None:
            raise InvalidMissionMetadata(f"metadata field: {key} does not exist")
        ret[metadata_type_id] = val
    return ret


def _update_mission(
    client: AuthenticatedClient, mission_id: UUID, *, metadata: Dict[str, str]
) -> None:
    tags_dct = _get_tags_map(client, metadata)
    payload = {
        "missionUUID": str(mission_id),
        "tags": {str(k): v for k, v in tags_dct.items()},
    }
    resp = client.post(UPDATE_MISSION, json=payload)

    if resp.status_code == 404:
        raise MissionNotFound
    if resp.status_code == 403:
        raise AccessDenied(f"cannot update mission: {mission_id}")

    resp.raise_for_status()


def _update_project(
    client: AuthenticatedClient,
    project_id: UUID,
    *,
    description: Optional[str] = None,
    new_name: Optional[str] = None,
) -> None:
    if description is None and new_name is None:
        raise ValueError("either description or new_name must be provided")

    body = {}
    if description is not None:
        body["description"] = description
    if new_name is not None:
        body["name"] = new_name
    resp = client.put(f"{UPDATE_PROJECT}/{project_id}", json=body)
    resp.raise_for_status()


def _get_api_version() -> Tuple[int, int, int]:
    config = get_config()
    client = httpx.Client()

    resp = client.get(f"{config.endpoint.api}{GET_STATUS}")
    vers = resp.headers["kleinkram-version"].split(".")

    return tuple(map(int, vers))  # type: ignore


def _claim_admin(client: AuthenticatedClient) -> None:
    """\
    the first user on the system could call this
    """
    response = client.post(CLAIM_ADMIN)
    response.raise_for_status()
    return


FILE_DELETE_MANY = "/files/deleteMultiple"


def _delete_files(
    client: AuthenticatedClient, file_ids: Sequence[UUID], mission_id: UUID
) -> None:
    payload = {
        "uuids": [str(file_id) for file_id in file_ids],
        "missionUUID": str(mission_id),
    }
    resp = client.post(FILE_DELETE_MANY, json=payload)
    resp.raise_for_status()


MISSION_DELETE_ONE = "/missions/{}"


def _delete_mission(client: AuthenticatedClient, mission_id: UUID) -> None:
    resp = client.delete(MISSION_DELETE_ONE.format(mission_id))

    # 409 is returned if the mission has files
    # 403 is returned if the mission does not exist / user cant delete

    resp.raise_for_status()


PROJECT_DELETE_ONE = "/projects/{}"


def _delete_project(client: AuthenticatedClient, project_id: UUID) -> None:
    resp = client.delete(PROJECT_DELETE_ONE.format(project_id))
    resp.raise_for_status()



from __future__ import annotations

from enum import Enum
from typing import Any
from typing import Dict
from typing import Generator
from typing import List
from typing import Mapping
from typing import Optional
from typing import cast

from kleinkram.api.client import AuthenticatedClient

DataPage = Dict[str, Any]


PAGE_SIZE = 128
SKIP = "skip"
TAKE = "take"


def paginated_request(
    client: AuthenticatedClient,
    endpoint: str,
    params: Optional[Mapping[str, Any]] = None,
    max_entries: Optional[int] = None,
    page_size: int = PAGE_SIZE,
) -> Generator[DataPage, None, None]:
    total_entries_count = 0

    params = dict(params or {})

    params[TAKE] = page_size
    params[SKIP] = 0

    while True:
        resp = client.get(endpoint, params=params)
        resp.raise_for_status()  # TODO: this is fine for now

        paged_data = resp.json()
        data_page = cast(List[DataPage], paged_data["data"])

        for entry in data_page:
            total_entries_count += 1
            yield entry
            if max_entries is not None and max_entries <= total_entries_count:
                return

        count = cast(int, paged_data["count"])
        skip = cast(int, paged_data["skip"])
        take = cast(int, paged_data["take"])

        if count - skip - take <= 0:
            return

        params[SKIP] = total_entries_count



from __future__ import annotations

LOGIN_MESSAGE = "Please login using `klein login`."


class ParsingError(Exception): ...


class InvalidFileQuery(Exception): ...


class InvalidMissionQuery(Exception): ...


class InvalidProjectQuery(Exception): ...


class MissionExists(Exception): ...


class ProjectExists(Exception): ...


class MissionNotFound(Exception): ...


class ProjectNotFound(Exception): ...


class FileNotFound(Exception): ...


class AccessDenied(Exception): ...


class NotAuthenticated(Exception):
    def __init__(self) -> None:
        super().__init__(LOGIN_MESSAGE)


class InvalidCLIVersion(Exception): ...


class FileTypeNotSupported(Exception): ...


class FileNameNotSupported(Exception): ...


class InvalidMissionMetadata(Exception): ...



"""\
this file contains wrappers around core functionality

these functions are meant to be exposed to the user, they
accept a more diverse set of arguments and handle the
conversion to the internal representation
"""

from __future__ import annotations

from pathlib import Path
from typing import Collection
from typing import Dict
from typing import List
from typing import Literal
from typing import Optional
from typing import Sequence
from typing import overload

import kleinkram.api.routes
import kleinkram.core
import kleinkram.utils
from kleinkram.api.client import AuthenticatedClient
from kleinkram.api.query import FileQuery
from kleinkram.api.query import MissionQuery
from kleinkram.api.query import ProjectQuery
from kleinkram.errors import FileNameNotSupported
from kleinkram.models import File
from kleinkram.models import Mission
from kleinkram.models import Project
from kleinkram.types import IdLike
from kleinkram.types import PathLike
from kleinkram.utils import parse_path_like
from kleinkram.utils import parse_uuid_like
from kleinkram.utils import singleton_list


def _args_to_project_query(
    project_names: Optional[Sequence[str]] = None,
    project_ids: Optional[Sequence[IdLike]] = None,
) -> ProjectQuery:
    return ProjectQuery(
        ids=[parse_uuid_like(_id) for _id in project_ids or []],
        patterns=list(project_names or []),
    )


def _args_to_mission_query(
    mission_names: Optional[Sequence[str]] = None,
    mission_ids: Optional[Sequence[IdLike]] = None,
    project_names: Optional[Sequence[str]] = None,
    project_ids: Optional[Sequence[IdLike]] = None,
) -> MissionQuery:
    return MissionQuery(
        ids=[parse_uuid_like(_id) for _id in mission_ids or []],
        patterns=list(mission_names or []),
        project_query=_args_to_project_query(
            project_names=project_names, project_ids=project_ids
        ),
    )


def _args_to_file_query(
    file_names: Optional[Sequence[str]] = None,
    file_ids: Optional[Sequence[IdLike]] = None,
    mission_names: Optional[Sequence[str]] = None,
    mission_ids: Optional[Sequence[IdLike]] = None,
    project_names: Optional[Sequence[str]] = None,
    project_ids: Optional[Sequence[IdLike]] = None,
) -> FileQuery:
    return FileQuery(
        ids=[parse_uuid_like(_id) for _id in file_ids or []],
        patterns=list(file_names or []),
        mission_query=_args_to_mission_query(
            mission_names=mission_names,
            mission_ids=mission_ids,
            project_names=project_names,
            project_ids=project_ids,
        ),
    )


def download(
    *,
    file_ids: Optional[Sequence[IdLike]] = None,
    file_names: Optional[Sequence[str]] = None,
    mission_ids: Optional[Sequence[IdLike]] = None,
    mission_names: Optional[Sequence[str]] = None,
    project_ids: Optional[Sequence[IdLike]] = None,
    project_names: Optional[Sequence[str]] = None,
    dest: PathLike,
    nested: bool = False,
    overwrite: bool = False,
    verbose: bool = False,
) -> None:
    query = _args_to_file_query(
        file_names=file_names,
        file_ids=file_ids,
        mission_names=mission_names,
        mission_ids=mission_ids,
        project_names=project_names,
        project_ids=project_ids,
    )
    client = AuthenticatedClient()
    kleinkram.core.download(
        client=client,
        query=query,
        base_dir=parse_path_like(dest),
        nested=nested,
        overwrite=overwrite,
        verbose=verbose,
    )


def list_files(
    *,
    file_ids: Optional[Sequence[IdLike]] = None,
    file_names: Optional[Sequence[str]] = None,
    mission_ids: Optional[Sequence[IdLike]] = None,
    mission_names: Optional[Sequence[str]] = None,
    project_ids: Optional[Sequence[IdLike]] = None,
    project_names: Optional[Sequence[str]] = None,
) -> List[File]:
    query = _args_to_file_query(
        file_names=file_names,
        file_ids=file_ids,
        mission_names=mission_names,
        mission_ids=mission_ids,
        project_names=project_names,
        project_ids=project_ids,
    )
    client = AuthenticatedClient()
    return list(kleinkram.api.routes.get_files(client, query))


def list_missions(
    *,
    mission_ids: Optional[Sequence[IdLike]] = None,
    mission_names: Optional[Sequence[str]] = None,
    project_ids: Optional[Sequence[IdLike]] = None,
    project_names: Optional[Sequence[str]] = None,
) -> List[Mission]:
    query = _args_to_mission_query(
        mission_names=mission_names,
        mission_ids=mission_ids,
        project_names=project_names,
        project_ids=project_ids,
    )
    client = AuthenticatedClient()
    return list(kleinkram.api.routes.get_missions(client, query))


def list_projects(
    *,
    project_ids: Optional[Sequence[IdLike]] = None,
    project_names: Optional[Sequence[str]] = None,
) -> List[Project]:
    query = _args_to_project_query(
        project_names=project_names,
        project_ids=project_ids,
    )
    client = AuthenticatedClient()
    return list(kleinkram.api.routes.get_projects(client, query))


@overload
def upload(
    *,
    mission_name: str,
    project_name: str,
    files: Sequence[PathLike],
    create: bool = False,
    fix_filenames: bool = False,
    metadata: Optional[Dict[str, str]] = None,
    ignore_missing_metadata: bool = False,
    verbose: bool = False,
) -> None: ...


@overload
def upload(
    *,
    mission_id: IdLike,
    files: Sequence[PathLike],
    create: Literal[False] = False,
    fix_filenames: bool = False,
    verbose: bool = False,
) -> None: ...


@overload
def upload(
    *,
    mission_name: str,
    project_id: IdLike,
    files: Sequence[PathLike],
    create: bool = False,
    fix_filenames: bool = False,
    metadata: Optional[Dict[str, str]] = None,
    ignore_missing_metadata: bool = False,
    verbose: bool = False,
) -> None: ...


def upload(
    *,
    mission_name: Optional[str] = None,
    mission_id: Optional[IdLike] = None,
    project_name: Optional[str] = None,
    project_id: Optional[IdLike] = None,
    files: Sequence[PathLike],
    create: bool = False,
    fix_filenames: bool = False,
    metadata: Optional[Dict[str, str]] = None,
    ignore_missing_metadata: bool = False,
    verbose: bool = False,
) -> None:
    parsed_file_paths = [parse_path_like(f) for f in files]
    if not fix_filenames:
        for file in parsed_file_paths:
            if not kleinkram.utils.check_filename_is_sanatized(file.stem):
                print(file.name)
                raise FileNameNotSupported(
                    f"only `{''.join(kleinkram.utils.INTERNAL_ALLOWED_CHARS)}` are "
                    f"allowed in filenames and at most 50 chars: {file}"
                )

    query = _args_to_mission_query(
        mission_names=singleton_list(mission_name),
        mission_ids=singleton_list(mission_id),
        project_names=singleton_list(project_name),
        project_ids=singleton_list(project_id),
    )
    client = AuthenticatedClient()
    kleinkram.core.upload(
        client=client,
        query=query,
        file_paths=parsed_file_paths,
        create=create,
        metadata=metadata,
        ignore_missing_metadata=ignore_missing_metadata,
        verbose=verbose,
    )


@overload
def verify(
    *,
    mission_name: str,
    project_name: str,
    files: Sequence[PathLike],
    verbose: bool = False,
) -> Dict[Path, kleinkram.core.FileVerificationStatus]: ...


@overload
def verify(
    *,
    mission_name: str,
    project_id: IdLike,
    files: Sequence[PathLike],
    verbose: bool = False,
) -> Dict[Path, kleinkram.core.FileVerificationStatus]: ...


@overload
def verify(
    *,
    mission_id: IdLike,
    files: Sequence[PathLike],
    verbose: bool = False,
) -> Dict[Path, kleinkram.core.FileVerificationStatus]: ...


def verify(
    *,
    mission_name: Optional[str] = None,
    mission_id: Optional[IdLike] = None,
    project_name: Optional[str] = None,
    project_id: Optional[IdLike] = None,
    files: Sequence[PathLike],
    skip_hash: bool = False,
    verbose: bool = False,
) -> Dict[Path, kleinkram.core.FileVerificationStatus]:
    query = _args_to_mission_query(
        mission_names=singleton_list(mission_name),
        mission_ids=singleton_list(mission_id),
        project_names=singleton_list(project_name),
        project_ids=singleton_list(project_id),
    )
    return kleinkram.core.verify(
        client=AuthenticatedClient(),
        query=query,
        file_paths=[parse_path_like(f) for f in files],
        skip_hash=skip_hash,
        verbose=verbose,
    )


def create_mission(
    mission_name: str,
    project_id: IdLike,
    metadata: Dict[str, str],
    ignore_missing_metadata: bool = False,
) -> None:
    kleinkram.api.routes._create_mission(
        AuthenticatedClient(),
        parse_uuid_like(project_id),
        mission_name,
        metadata=metadata,
        ignore_missing_tags=ignore_missing_metadata,
    )


def create_project(project_name: str, description: str) -> None:
    kleinkram.api.routes._create_project(
        AuthenticatedClient(), project_name, description
    )


def update_file(file_id: IdLike) -> None:
    kleinkram.core.update_file(
        client=AuthenticatedClient(), file_id=parse_uuid_like(file_id)
    )


def update_mission(mission_id: IdLike, metadata: Dict[str, str]) -> None:
    kleinkram.core.update_mission(
        client=AuthenticatedClient(),
        mission_id=parse_uuid_like(mission_id),
        metadata=metadata,
    )


def update_project(project_id: IdLike, description: Optional[str] = None) -> None:
    kleinkram.core.update_project(
        client=AuthenticatedClient(),
        project_id=parse_uuid_like(project_id),
        description=description,
    )


def delete_files(file_ids: Collection[IdLike]) -> None:
    """\
    delete multiple files by their ids
    """
    kleinkram.core.delete_files(
        client=AuthenticatedClient(),
        file_ids=[parse_uuid_like(_id) for _id in file_ids],
    )


def delete_file(file_id: IdLike) -> None:
    """\
    delete a single file by id
    """
    file = kleinkram.api.routes.get_file(
        AuthenticatedClient(), FileQuery(ids=[parse_uuid_like(file_id)])
    )
    kleinkram.api.routes._delete_files(
        AuthenticatedClient(), file_ids=[file.id], mission_id=file.mission_id
    )


def delete_mission(mission_id: IdLike) -> None:
    kleinkram.core.delete_mission(
        client=AuthenticatedClient(), mission_id=parse_uuid_like(mission_id)
    )


def delete_project(project_id: IdLike) -> None:
    kleinkram.core.delete_project(
        client=AuthenticatedClient(), project_id=parse_uuid_like(project_id)
    )


def get_file(file_id: IdLike) -> File:
    """\
    get a file by its id
    """
    return kleinkram.api.routes.get_file(
        AuthenticatedClient(), FileQuery(ids=[parse_uuid_like(file_id)])
    )


def get_mission(mission_id: IdLike) -> Mission:
    """\
    get a mission by its id
    """
    return kleinkram.api.routes.get_mission(
        AuthenticatedClient(), MissionQuery(ids=[parse_uuid_like(mission_id)])
    )


def get_project(project_id: IdLike) -> Project:
    """\
    get a project by its id
    """
    return kleinkram.api.routes.get_project(
        AuthenticatedClient(), ProjectQuery(ids=[parse_uuid_like(project_id)])
    )



from __future__ import annotations

from pathlib import Path
from typing import Optional

import typer

import kleinkram.api.routes
import kleinkram.core
from kleinkram.api.client import AuthenticatedClient
from kleinkram.api.query import MissionQuery
from kleinkram.api.query import ProjectQuery
from kleinkram.api.routes import get_mission
from kleinkram.api.routes import get_project
from kleinkram.config import get_shared_state
from kleinkram.printing import print_mission_info
from kleinkram.utils import load_metadata
from kleinkram.utils import split_args

CREATE_HELP = "create a mission"
UPDATE_HELP = "update a mission"
DELETE_HELP = "delete a mission"
INFO_HELP = "get information about a mission"
NOT_IMPLEMENTED_YET = """\
Not implemented yet, open an issue if you want specific functionality
"""

mission_typer = typer.Typer(
    no_args_is_help=True, context_settings={"help_option_names": ["-h", "--help"]}
)


@mission_typer.command(help=CREATE_HELP)
def create(
    project: str = typer.Option(..., "--project", "-p", help="project id or name"),
    mission_name: str = typer.Option(..., "--mission", "-m", help="mission name"),
    metadata: Optional[str] = typer.Option(
        None, help="path to metadata file (json or yaml)"
    ),
    ignore_missing_tags: bool = typer.Option(False, help="ignore mission tags"),
) -> None:
    project_ids, project_patterns = split_args([project] if project else [])
    project_query = ProjectQuery(ids=project_ids, patterns=project_patterns)

    metadata_dct = load_metadata(Path(metadata)) if metadata else {}  # noqa

    client = AuthenticatedClient()
    project_id = get_project(client, project_query).id
    mission_id = kleinkram.api.routes._create_mission(
        client,
        project_id,
        mission_name,
        metadata=metadata_dct,
        ignore_missing_tags=ignore_missing_tags,
    )

    mission_parsed = get_mission(client, MissionQuery(ids=[mission_id]))
    print_mission_info(mission_parsed, pprint=get_shared_state().verbose)


@mission_typer.command(help=INFO_HELP)
def info(
    project: Optional[str] = typer.Option(
        None, "--project", "-p", help="project id or name"
    ),
    mission: str = typer.Option(..., "--mission", "-m", help="mission id or name"),
) -> None:
    mission_ids, mission_patterns = split_args([mission])
    project_ids, project_patterns = split_args([project] if project else [])

    project_query = ProjectQuery(ids=project_ids, patterns=project_patterns)
    mission_query = MissionQuery(
        ids=mission_ids,
        patterns=mission_patterns,
        project_query=project_query,
    )

    client = AuthenticatedClient()
    mission_parsed = get_mission(client, mission_query)
    print_mission_info(mission_parsed, pprint=get_shared_state().verbose)


@mission_typer.command(help=UPDATE_HELP)
def update(
    project: Optional[str] = typer.Option(
        None, "--project", "-p", help="project id or name"
    ),
    mission: str = typer.Option(..., "--mission", "-m", help="mission id or name"),
    metadata: str = typer.Option(help="path to metadata file (json or yaml)"),
) -> None:
    mission_ids, mission_patterns = split_args([mission])
    project_ids, project_patterns = split_args([project] if project else [])

    project_query = ProjectQuery(ids=project_ids, patterns=project_patterns)
    mission_query = MissionQuery(
        ids=mission_ids,
        patterns=mission_patterns,
        project_query=project_query,
    )

    metadata_dct = load_metadata(Path(metadata))

    client = AuthenticatedClient()
    mission_id = get_mission(client, mission_query).id
    kleinkram.core.update_mission(
        client=client, mission_id=mission_id, metadata=metadata_dct
    )

    mission_parsed = get_mission(client, mission_query)
    print_mission_info(mission_parsed, pprint=get_shared_state().verbose)


@mission_typer.command(help=DELETE_HELP)
def delete(
    project: Optional[str] = typer.Option(
        None, "--project", "-p", help="project id or name"
    ),
    mission: str = typer.Option(..., "--mission", "-m", help="mission id or name"),
    confirm: bool = typer.Option(
        False, "--confirm", "-y", "--yes", help="confirm deletion"
    ),
) -> None:
    if not confirm:
        typer.confirm(f"delete {project} {mission}", abort=True)

    project_ids, project_patterns = split_args([project] if project else [])
    project_query = ProjectQuery(ids=project_ids, patterns=project_patterns)

    mission_ids, mission_patterns = split_args([mission])
    mission_query = MissionQuery(
        ids=mission_ids,
        patterns=mission_patterns,
        project_query=project_query,
    )

    client = AuthenticatedClient()
    mission_parsed = get_mission(client, mission_query)
    kleinkram.core.delete_mission(client=client, mission_id=mission_parsed.id)


@mission_typer.command(help=NOT_IMPLEMENTED_YET)
def prune(
    project: Optional[str] = typer.Option(
        None, "--project", "-p", help="project id or name"
    ),
    mission: str = typer.Option(..., "--mission", "-m", help="mission id or name"),
) -> None:
    """\
    delete files with bad file states, e.g. missing not uploaded corrupted etc.
    TODO: open for suggestions what this should do
    """

    raise NotImplementedError("Not implemented yet")



from __future__ import annotations

import pytest

from kleinkram import list_files
from kleinkram import list_missions
from kleinkram import list_projects
from testing.backend_fixtures import DATA_FILES
from testing.backend_fixtures import PROJECT_DESCRIPTION


@pytest.mark.slow
def test_project_fixture(project):
    assert list_projects(project_ids=[project.id])[0].id == project.id
    assert project.description == PROJECT_DESCRIPTION


@pytest.mark.slow
def test_mission_fixture(mission, project):
    assert mission.project_id == project.id
    assert list_missions(mission_ids=[mission.id])[0].id == mission.id

    files = list_files(mission_ids=[mission.id])

    assert set([file.name for file in files if file.name.endswith(".bag")]) == set(
        [file.name for file in DATA_FILES]
    )


@pytest.mark.slow
def test_empty_mission_fixture(empty_mission, project):
    assert empty_mission.project_id == project.id
    assert list_missions(mission_ids=[empty_mission.id])[0].id == empty_mission.id
    assert not list_files(mission_ids=[empty_mission.id])
